#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys
import os

from deeptools import parserCommon, bamHandler, utilities
from deeptools.mapReduce import mapReduce
from deeptoolsintervals import GTF
import numpy as np
import pandas as pd

# plotting
import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("Agg")
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["svg.fonttype"] = "none"

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei.Utilities import *
from sincei import ParserCommon


def parseArguments():
    io_args = ParserCommon.inputOutputOptions(opts=["bamfile", "whitelist", "outFile"], requiredOpts=["bamfile"])
    bam_args = ParserCommon.bamOptions(
        suppress_args=["labels", "smartLabels", "distanceBetweenBins", "region"],
        default_opts={"binSize": 100000},
    )
    other_args = ParserCommon.otherOptions()

    parser = argparse.ArgumentParser(
        parents=[io_args, get_args(), bam_args, other_args],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
        This tool identifies barcodes present in a BAM files and produces a list. You can optionally filter these
        barcodes by matching them to a whitelist or based on total counts.
        """,
        usage="Example usage: scFilterBarcodes -b sample.bam -w whitelist.txt > barcodes_detected.txt",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)
    general = parser.add_argument_group("Counting Options")

    general.add_argument(
        "--minHammingDist",
        "-d",
        help="Minimum hamming distance to match the barcode in whitelist. Note that increasing the "
        "hamming distance really slows down the barcode detection process.",
        metavar="INT",
        type=int,
        default=0,
        required=False,
    )

    general.add_argument(
        "--minCount",
        "-mc",
        help="Minimum no. of bins with non-zero counts, in order to report a barcode. Note that this number would range "
        "from 0 to genome size/binSize. ",
        metavar="INT",
        type=int,
        default=0,
        required=False,
    )

    general.add_argument(
        "--minMappingQuality",
        "-mq",
        metavar="INT",
        help="If set, only reads that have a mapping " "quality score of at least this are " "considered.",
        type=int,
    )

    general.add_argument(
        "--rankPlot",
        "-rp",
        type=parserCommon.writableFile,
        help='The output file name to plot the ranked counts per barcode (similar to the "knee plot",'
        "but counts in this case would be the number of non-zero bins)",
    )

    return parser


##
def ham_dist(s1, s2):
    r"""get hamming dist between 2 sequences

    Parameters
    ----------
    s1, s2 : string, sequences

    Returns
    ----------
    integer hamming distance
    """
    if len(s1) != len(s2):
        raise ValueError("Undefined")
    return sum(ch1 != ch2 for ch1, ch2 in zip(s1, s2))


def getFiltered_worker(arglist):
    r"""get a set of filtered barcodes based on the provided criteria

    Parameters
    ----------
    arglist : list of filtering arguments

    Returns
    ----------
    BCset: a set of barcodes
    """
    chrom, start, end, args = arglist
    # Fix the bounds
    if end <= start:
        end = start + 1
    ## open blacklist file
    blackList = None
    if args.blackListFileName is not None:
        blackList = GTF(args.blackListFileName)

    fh = bamHandler.openBam(args.bamfile)
    chromUse = utilities.mungeChromosome(chrom, fh.references)

    BCset = set()
    for read in fh.fetch(chromUse, start, end):
        if read.pos < start:
            # ensure that we never double count (in case distanceBetweenBins == 0)
            continue

        if read.flag & 4:
            # Ignore unmapped reads, they were counted already
            continue

        if args.minMappingQuality and read.mapq < args.minMappingQuality:
            continue

        ## reads in blacklisted regions
        if blackList and blackList.findOverlaps(
            chrom,
            read.reference_start,
            read.reference_start + read.infer_query_length(always=False) - 1,
        ):
            continue

        ## get barcode and append to set
        try:
            bc = read.get_tag(args.cellTag)
        except KeyError:
            continue
        if args.whitelist:
            # match barcode to whitelist
            if args.minHammingDist == 0:
                if bc in args.whitelist:
                    BCset.add(bc)
            else:
                try:
                    hamdist = [ham_dist(x, bc) for x in args.whitelist]
                    if min(hamdist) <= args.minHammingDist:
                        BCset.add(bc)
                except ValueError:
                    continue
        else:
            BCset.add(bc)
    fh.close()

    # return the set with barcodes detected
    return BCset


#
def count_occurrences(res):
    r"""count occurrences of elements (barcodes) in a list of sets

    Parameters
    ----------
    res : list of sets of barcodes

    Returns
    ----------
    counts
    """
    barcodes = set.union(*res)
    counts = {barcode: 0 for barcode in barcodes}
    for set_ in res:
        for barcode in set_:
            counts[barcode] += 1
    return counts


def main(args=None):
    args = parseArguments().parse_args(args)
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    # open barcode file and read the content in a list
    if args.whitelist:
        with open(args.whitelist, "r") as f:
            barcodes = f.read().splitlines()
        args.whitelist = barcodes

    bhs = bamHandler.openBam(args.bamfile, returnStats=True, nThreads=args.numberOfProcessors)[0]
    chrom_sizes = list(zip(bhs.references, bhs.lengths))
    bhs.close()

    # Get the remaining metrics
    res = mapReduce(
        [args],
        getFiltered_worker,
        chrom_sizes,
        genomeChunkLength=args.binSize,
        blackListFileName=args.blackListFileName,
        numberOfProcessors=args.numberOfProcessors,
        verbose=args.verbose,
    )
    ## res, should be a list of sets
    df = pd.DataFrame.from_dict(count_occurrences(res), orient="index", columns=["count"]).reset_index()
    df.columns = ["barcode", "count"]
    df["selected"] = True

    if args.minCount:
        if args.minCount > len(res):
            print("minCount bigger than No. of bins. Reducing to maximum")
            args.minCount = len(res)
        final_set = df.loc[df["count"] >= args.minCount]["barcode"].to_list()
        df.loc[df["count"] < args.minCount, "selected"] = False
    else:
        final_set = df["barcode"].tolist()

    # convert count to log10
    if args.rankPlot:
        df["count_log10"] = np.log10(df["count"])
        df["count_rank"] = df["count"].rank(method="min", ascending=False)
        xrange = np.arange(0, np.round(max(df["count_rank"]), -3), np.round(int(max(df["count_rank"]) / 10), -3))
        yrange = np.arange(
            np.round(min(df["count_log10"]), 2),
            np.round(max(df["count_log10"]), 2),
            np.round(max(df["count_log10"]) / 10, 2),
        )

        fig, ax = plt.subplots()
        plt.plot(
            "count_rank",
            "count_log10",
            data=df,
            linestyle="none",
            marker="o",
            color="grey",
            markersize=0.5,
        )
        ax.set_xlabel("Barcode Rank", fontsize=12)
        ax.set_ylabel("No. of nonzero bins (log10)", fontsize=12)
        ax.set_title("Ranked counts (#bins) for detected Barcodes", fontsize=13)
        plt.xticks(xrange, fontsize=10)
        plt.yticks(yrange, fontsize=10)

        # Annotation
        if args.minCount:
            plt.axhline(np.log10(args.minCount), color="r")

        fig.tight_layout()
        plt.savefig(
            args.rankPlot,
            dpi="figure",
            format=None,
            metadata=None,
            pad_inches=0.2,
            facecolor="auto",
            edgecolor="auto",
            backend=None,
        )

    if args.outFile is None:
        of = sys.stdout
    else:
        of = open(args.outFile, "w")

    df.to_csv(of, sep="\t")

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import numpy as np
import pandas as pd

# logs
import sys
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

# plotting
import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("Agg")
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["svg.fonttype"] = "none"

# single-cell stuff
import anndata as ad
import scanpy as sc


from sincei import ParserCommon
from sincei.TopicModels import TOPICMODEL

from sincei.GLMPCA import EXPONENTIAL_FAMILY_DICT  # GLMPCA


def parseArguments():
    io_args = ParserCommon.inputOutputOptions(opts=["h5adfile", "outFile"], requiredOpts=["outFile"])
    plot_args = ParserCommon.plotOptions()
    other_args = ParserCommon.otherOptions()

    parser = argparse.ArgumentParser(
        parents=[io_args, get_args(), plot_args, other_args],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,  # argparse.RawDescriptionHelpFormatter,
        description="""
        This tool clusters the cells based on the input count matrix (output of scCountReads) and performs dimensionality reduction,
        community detection and 2D projection (UMAP) of the cells. The result is an updated h5ad object, and (optionally) a plot file
        and a tsv file with UMAP coordinates and corresponding cluster id for each barcode.
        """,
        usage="Example usage: scClusterCells -i cellCounts.h5ad -o clustered.h5ad -op <umap_prefix>.png  > log.txt",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)

    general = parser.add_argument_group("Clustering Options")
    general.add_argument(
        "--outFileUMAP",
        "-op",
        type=str,
        required=False,
        help="The output plot file (for UMAP). If you specify this option, another file with the same "
        "prefix (and .txt extention) is also created with the raw UMAP coordinates.",
    )

    general.add_argument(
        "--outFileTrainedModel",
        "-om",
        type=argparse.FileType("w"),
        required=False,
        help="The output file for the trained LSI model. The saved model can be used later to embed/compare new cells "
        "to the existing cluster of cells.",
    )

    general.add_argument(
        "--method",
        "-m",
        type=str,
        choices=["logPCA", "LSA", "LDA", "glmPCA"],
        default="LSA",
        help="The dimensionality reduction method for clustering. (Default: %(default)s)",
    )

    general.add_argument(
        "--glmPCAfamily",
        "-gf",
        type=str,
        choices=EXPONENTIAL_FAMILY_DICT.keys(),
        default="poisson",
        help="The choice of exponential family distribution to use for glmPCA method. (Default: %(default)s)",
    )

    general.add_argument(
        "--binarize",
        action="store_true",
        help="Binarize the counts per region before dimensionality reduction (only for LSA/LDA)",
    )

    general.add_argument(
        "--nPrinComps",
        "-n",
        default=20,
        type=int,
        help="Number of principle components to reduce the dimensionality to. "
        "Use higher number for samples with more expected heterogenity. (Default: %(default)s)",
    )

    general.add_argument(
        "--nNeighbors",
        "-nk",
        default=30,
        type=int,
        help="Number of nearest neighbours to consider for clustering and UMAP. This number should be chosen considering "
        "the total number of cells and expected number of clusters. Smaller number will lead to more fragmented clusters. "
        "(Default: %(default)s)",
    )

    general.add_argument(
        "--clusterResolution",
        "-cr",
        default=1.0,
        type=float,
        help="Resolution parameter for clustering. Values lower than 1.0 would result in less clusters, "
        "while higher values lead to splitting of clusters. In most cases, the optimum value would be between "
        "0.8 and 1.2. (Default: %(default)s)",
    )

    return parser


def main(args=None):
    args = parseArguments().parse_args(args)
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    try:
        adata = sc.read_h5ad(args.input)
    except:
        sys.stderr.write("\n Error: Input file can not be read (doesn't appear to be a valid anndata object) \n")
        exit()

    if args.method == "logPCA":
        ## log1p+PCA using scanpy
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.pca(adata, args.nPrinComps)

    elif args.method == "LSA":
        ## LSA using gensim
        model_object = TOPICMODEL(
            adata,
            binarize=args.binarize,
            n_topics=args.nPrinComps,
            smart_code="lfu",
        )
        model_object.runLSA()
        cell_topic = model_object.get_cell_topic(pop_sparse_cells=True)
        ## update the anndata object, drop cells which are not in the anndata, drop 1st PC
        adata = adata[cell_topic.index]
        adata.obsm["X_pca"] = np.asarray(cell_topic.iloc[:, 1 : args.nPrinComps])

    elif args.method == "LDA":
        ## LDA using gensim
        model_object = TOPICMODEL(
            adata,
            binarize=args.binarize,
            n_topics=args.nPrinComps,
            n_passes=2,
            n_workers=4,
        )
        model_object.runLDA()
        cell_topic = model_object.get_cell_topic(pop_sparse_cells=True)
        ## update the anndata object, drop cells which are not in the anndata, drop 1st PC
        adata = adata[cell_topic.index]
        adata.obsm["X_pca"] = np.asarray(cell_topic.iloc[:, 1 : args.nPrinComps])

    elif args.method == "glmPCA":
        # import glmPCA (not imported on top due to special optional import of mctorch)
        from sincei.GLMPCA import GLMPCA

        ## glmPCA using mctorch
        model_object = GLMPCA(
            n_pc=args.nPrinComps,
            family=args.glmPCAfamily,
        )
        model_object.fit(adata)
        cell_pcs = model_object.saturated_loadings_.detach().numpy()

        ## update the anndata object
        adata.obsm["X_pca"] = np.asarray(cell_pcs)

    sc.pp.neighbors(adata, use_rep="X_pca", n_neighbors=args.nNeighbors)
    sc.tl.leiden(adata, resolution=args.clusterResolution)
    sc.tl.paga(adata)
    sc.pl.paga(adata, plot=False, threshold=0.1)
    sc.tl.umap(adata, min_dist=0.1, spread=5, init_pos="paga")

    adata.write_h5ad(args.outFile)

    if args.outFileUMAP:
        ## plot UMAP
        fig = sc.pl.umap(adata, color="leiden", legend_loc="on data", return_fig=True, show=False)
        fig.savefig(args.outFileUMAP, dpi=300, format=args.plotFileFormat)
        prefix = args.outFileUMAP.split(".")[0]
        umap_lsi = pd.DataFrame(adata.obsm["X_umap"], columns=["UMAP1", "UMAP2"], index=adata.obs.index)
        umap_lsi["cluster"] = adata.obs["leiden"]
        umap_lsi.to_csv(prefix + ".tsv", sep="\t", index_label="Cell_ID")
        # plt.rcParams['font.size'] = 8.0
        # convert cm values to inches
        # fig = plt.figure(figsize=(args.plotWidth / 2.54, args.plotHeight / 2.54))
        # fig.suptitle('LSA-UMAP', y=(1 - (0.06 / args.plotHeight)))
        # plt.scatter(umap_lsi.UMAP1, umap_lsi.UMAP2, s=5, alpha = 0.8, c=[sns.color_palette()[x] for x in list(umap_lsi.cluster)])
        # plt.tight_layout()
        # plt.savefig(args.outFileUMAP, dpi=200, format=args.plotFileFormat)
        # plt.close()

    # save if asked
    if args.outFileTrainedModel:
        model_object.lsi_model.save(args.outFileTrainedModel)
    # if args.outGraph:
    # 'The output file for the Graph object (lgl format) which can be used for further clustering/integration.'
    #   graph.write_lgl(args.outGraph)

    return 0



def read_mtx(prefix):
    r"""Reads a matrix from a mtx file

    Parameters
    ----------
    prefix : str
        prefix of the mtx file

    Returns
    -------
    mtx : scipy.sparse.csr_matrix
        matrix
    rownames : list
        list of row names
    colnames : list
        list of column names

    Examples
    --------

    >>> mtx, rownames, colnames = read_mtx("test.mtx")
    """
    mtx = io.mmread(prefix + ".counts.mtx")
    mtx = mtx.tocsr()
    with open(prefix + ".colnames.txt") as col:
        colnames = col.read().splitlines()
    col.close()
    with open(prefix + ".rownames.txt") as row:
        rownames = row.read().splitlines()
    row.close()

    return mtx, rownames, colnames


# from mtx
def preprocess_mtx(sparse_mtx, rownames, colnames, min_cell_sum, min_region_sum):
    r"""Preprocesses a sparse matrix for use with scanpy

    Parameters
    ----------
    sparse_mtx : scipy.sparse.csr_matrix
        Sparse matrix of read counts

    rownames : list
        List of row names

    colnames : list
        List of column names

    min_cell_sum : int
        Minimum number of reads per cell

    min_region_sum : int
        Minimum number of reads per region

    Returns
    -------
    scanpy.AnnData
        AnnData object with sparse matrix of read counts


    Examples
    --------

    >>> test = Tester()
    >>> sparse_mtx = test.sparse_mtx
    >>> rownames = test.rownames
    >>> colnames = test.colnames
    >>> adata = preprocess_mtx(sparse_mtx, rownames, colnames, min_cell_sum=1, min_region_sum=1)
    """

    from itertools import compress

    ## binarize
    nonzero_mask = np.array(sparse_mtx[sparse_mtx.nonzero()] > 1)[0]
    rows = sparse_mtx.nonzero()[0][nonzero_mask]
    cols = sparse_mtx.nonzero()[1][nonzero_mask]
    sparse_mtx[rows, cols] = 1

    ## filter low counts
    colmask = np.array(np.sum(sparse_mtx, axis=0) >= min_cell_sum)[0]
    rowmask = np.array(np.sum(sparse_mtx, axis=1) >= min_region_sum)
    rowmask = np.array([x[0] for x in rowmask])
    sparse_mtx = sparse_mtx[rowmask, :]
    sparse_mtx = sparse_mtx[:, colmask]

    ## create anndata
    row_subset = list(compress(rownames, rowmask))
    col_subset = list(compress(colnames, colmask))
    adata = anndata.AnnData(
        sparse_mtx.transpose(),
        obs=pd.DataFrame(col_subset),
        var=pd.DataFrame(row_subset),
    )

    return adata


# Imports for cluster_topic
# from sklearn.metrics import pairwise_distances
# from scanpy.neighbors import (
#     _compute_connectivities_umap,
#     _get_indices_distances_from_dense_matrix,
# )
# from scanpy._utils import get_igraph_from_adjacency
# import leidenalg as la
# import umap
# import pandas as pd


def cluster_topic(
    cell_topic,
    distance_metric="cosine",
    nk=30,
    resolution=1.0,
    connectivity_graph=True,
):
    r"""Cluster cells using the output of LSA_gensim or LDA_gensim using the Leiden algorithm.

    Parameters
    ----------
    cell_topic : pandas dataframe
        cell_topic dataframe with cell names as index and topic proportions as columns.
    distance_metric : str
        Distance metric to use. Default: 'cosine'
    nk : int
        Number of nearest neighbors to use. Default: 30
    resolution : float
        Resolution parameter for modularity algorithm. Default: 1.0
    connectivity_graph : bool
        Whether to use a connectivity graph or a distance graph. Default: True

    Returns
    -------
    umap_df : pandas dataframe
        UMAP embedding with cluster labels.
    G : igraph object
        Graph object.
    """

    # cluster on cell-topic dist
    _distances = pairwise_distances(cell_topic.iloc[:, 1:], metric=distance_metric)
    knn_indices, knn_distances = _get_indices_distances_from_dense_matrix(_distances, nk)
    distances, connectivities = _compute_connectivities_umap(knn_indices, knn_distances, _distances.shape[0], nk)

    if connectivity_graph:
        G = get_igraph_from_adjacency(connectivities, directed=True)
    else:
        G = get_igraph_from_adjacency(distances, directed=True)
    partition = la.find_partition(
        G,
        la.RBConfigurationVertexPartition,
        weights="weight",
        seed=42,
        resolution_parameter=resolution,
    )
    cell_topic["cluster"] = partition.membership

    # umap on cell-topic dist
    um = umap.UMAP(
        spread=5,
        min_dist=0.1,
        n_neighbors=nk,
        metric=distance_metric,
        init="random",
        random_state=42,
    )
    umfit = um.fit(cell_topic.iloc[:, 0 : (len(cell_topic.columns) - 1)])
    umap_df = pd.DataFrame(umfit.embedding_)
    umap_df.columns = ["UMAP1", "UMAP2"]
    umap_df["cluster"] = list(cell_topic.cluster)
    umap_df.index = cell_topic.index

    return umap_df, G


## LSA using scipy (older version)
# def runLSA(sparse_mtx, nPCs, scaleFactor):
# tf = (sparse_mtx.transpose() / sparse_mtx.sum(axis=0)).transpose()
#    tf = sparse_mtx / sparse_mtx.sum(axis=0)
#    tf = np.log1p(tf * scaleFactor)

#    idf = np.log1p(sparse_mtx.shape[1] / sparse_mtx.sum(axis=1))
#    tfidf = np.multiply(tf, idf)
#    svd = TruncatedSVD(n_components=nPCs)
#    pca = svd.fit(np.nan_to_num(tfidf, nan = 0.0))

#    return pca, tfidf

## for anndata
# def lsa_anndata(adata, n_pcs, scale_factor):
#   from scipy.sparse import issparse, coo_matrix, csr_matrix
#    mtx = sparse.csr_matrix(adata.X)
#    lsa_out, tfidf = runLSA(mtx.transpose(), n_pcs, scale_factor)

#    adata.X = np.squeeze(np.asarray(tfidf.transpose()))
#    adata.obsm['X_pca'] = lsa_out.components_.transpose()
#    adata.uns['pca_variance'] = lsa_out.explained_variance_
#    adata.uns['pca_variance_ratio'] = lsa_out.explained_variance_ratio_
#    adata.layers['raw_counts'] = mtx

#    return adata

# def UMAP_clustering(adata):

# make umap on PCA
#    lsa_out = adata.obsm['X_pca'].transpose()[1:, :]
#    reducer = umap.UMAP(n_neighbors=15, min_dist=0.1, spread=1.0, metric='euclidean', init = 'random')
#    embeddings = reducer.fit_transform(lsa_out.transpose())
#    adata.obsm['X_umap'] = embeddings

# louvain (from scanpy)
#    scp.pp.neighbors(adata)
#    scp.tl.louvain(adata)
#    cluster_id = [int(x) for x in adata.obs['louvain'].to_list()]

# return
#    return adata



import os
import sys
import shutil
import numpy as np
import pandas as pd

# deeptools modules
from deeptools import mapReduce
from deeptools.utilities import getCommonChrNames
from deeptools import bamHandler
from deeptools import utilities
from deeptools.writeBedGraph import bedGraphToBigWig, getGenomeChunkLength

# own modules
from sincei import ReadCounter as cr


def scaleCoverage(tile_coverage, args):
    r"""
    Return coverage per cluster as sum of cells.
    tileCoverage should be an list with only one element
    """
    return args["scaleFactor"] * tile_coverage


def writeBedGraph_wrapper(args):
    r"""
    Passes the arguments to writeBedGraph_worker.
    This is a step required given
    the constrains from the multiprocessing module.
    The args var, contains as first element the 'self' value
    from the WriteBedGraph object
    """
    return WriteBedGraph.writeBedGraph_worker(*args)


class WriteBedGraph(cr.CountReadsPerBin):
    r"""Reads bam files coverages and writes a bedgraph or bigwig file

    Extends the CountReadsPerBin object such that the coverage
    of bam files is writen to multiple bedgraph files at once.

    The bedgraph files are later merge into one and converted
    into a bigwig file if necessary.

    The constructor arguments are the same as for CountReadsPerBin. However,
    when calling the `run` method, the following parameters have
    to be passed

    Examples
    --------
    Given the following distribution of reads that cover 200 on
    a chromosome named '3R'::

          0                              100                           200
          |------------------------------------------------------------|
        A                                ===============
                                                        ===============


        B                 ===============               ===============
                                         ===============
                                                        ===============

    >>> import tempfile
    >>> test_path = os.path.dirname(os.path.abspath(__file__)) + "/test/test_data/"

    >>> outFile = tempfile.NamedTemporaryFile()
    >>> bam_file = test_path +  "testA.bam"

    For the example a simple scaling function is going to be used. This function
    takes the coverage found at each region and multiplies it to the scaling factor.
    In this case the scaling factor is 1.5

    >>> function_to_call = scaleCoverage
    >>> funcArgs = {'scaleFactor': 1.5}

    Restrict process to a region between positions 0 and 200 of chromosome 3R

    >>> region = '3R:0:200'

    Set up such that coverage is computed for consecutive bins of length 25 bp
    >>> bin_length = 25
    >>> step_size = 25

    >>> num_sample_sites = 0 #overruled by step_size
    >>> c = WriteBedGraph([bam_file], binLength=bin_length, region=region, stepSize=step_size)
    >>> c.run(function_to_call, funcArgs, outFile.name)
    >>> f = open(outFile.name, 'r')
    >>> f.readlines()
    ['3R\t0\t100\t0\n', '3R\t100\t200\t1.5\n']
    >>> f.close()
    >>> outFile.close()
    """

    def run(
        self,
        func_to_call,
        func_args,
        out_file_prefix,
        blackListFileName=None,
        format="bedgraph",
        smoothLength=0,
        normUsing=None,
    ):
        r"""
        Given a list of bamfiles, a function and a function arguments,
        this method writes a bedgraph file (or bigwig) file
        for a partition of the genome into tiles of given size
        and a value for each tile that corresponds to the given function
        and that is related to the coverage underlying the tile.

        Parameters
        ----------
        func_to_call : str
            function name to be called to convert the list of coverages computed
            for each bam file at each position into a single value.
        func_args : dict
            dict of arguments to pass to `func`. E.g. {'scaleFactor':1.0}
        out_file_prefix : str
            name of the file to save the resulting data.
        """
        # self.__dict__["smoothLength"] = smoothLength
        getStats = len(self.mappedList) < len(self.bamFilesList)
        bam_handles = []
        for x in self.bamFilesList:
            if getStats:
                bam, mapped, unmapped, stats = bamHandler.openBam(x, returnStats=True, nThreads=self.numberOfProcessors)
                self.mappedList.append(mapped)
                self.statsList.append(stats)
            else:
                bam = bamHandler.openBam(x)
            bam_handles.append(bam)

        genome_chunk_length = getGenomeChunkLength(bam_handles, self.binLength, self.mappedList)
        # check if both bam files correspond to the same species
        # by comparing the chromosome names:
        chrom_names_and_size, non_common = getCommonChrNames(bam_handles, verbose=False)

        if self.region:
            # in case a region is used, append the tilesize
            self.region += ":{}".format(self.binLength)

        for x in list(self.__dict__.keys()):
            if x in [
                "mappedList",
                "statsList",
                "barcodes",
                "clusterInfo",
                "groupLabels",
            ]:
                continue
            sys.stderr.write("{}: {}\n".format(x, self.__getattribute__(x)))

        # below we get the same ouput as in deeptools, except that the 3rd list
        # element contains multiple tmp file names, one tmp file per cluster
        res = mapReduce.mapReduce(
            [func_to_call, func_args],
            writeBedGraph_wrapper,
            chrom_names_and_size,
            self_=self,
            genomeChunkLength=genome_chunk_length,
            region=self.region,
            blackListFileName=blackListFileName,
            numberOfProcessors=self.numberOfProcessors,
        )

        # Determine the sorted order of the temp files
        chrom_order = dict()
        for i, _ in enumerate(chrom_names_and_size):
            chrom_order[_[0]] = i
        res = [[chrom_order[x[0]], x[1], x[2], x[3]] for x in res]
        res.sort()

        # write output for each group
        cluster_info = self.clusterInfo
        clusters = cluster_info["cluster"].unique().tolist()
        prefix = os.path.splitext(os.path.basename(out_file_prefix))[0]
        for cl in clusters:
            print("Writing output for group: {}".format(cl))
            if pd.isna(cl):
                continue
            # concatenate the coverages
            tmp_out = "/tmp/{}_{}.tmp".format(prefix, cl)
            out_file = open(tmp_out, "wb")
            for r in res:
                if r[3][cl]:
                    _foo = open(r[3][cl], "rb")
                    shutil.copyfileobj(_foo, out_file)
                    _foo.close()
                    os.remove(r[3][cl])
            out_file.close()

            ## read back and normalize
            cl_idx = cluster_info.index[pd.Series(cluster_info.cluster == cl)].tolist()
            nCells = float(len(cl_idx))
            out_file = pd.read_csv(tmp_out, sep="\t", index_col=None, header=None)
            # CPM norm
            if normUsing == "CPM":
                mil_reads_mapped = float(np.sum(out_file[3])) / 1e6
                if mil_reads_mapped < 0.00001:
                    sys.stderr.write(
                        "\n No or too few reads counted for group: {} ."
                        ". If this persists for all groups, please double-check that your barcodes"
                        " match between the groupInfo file and the BAM files and you specified the correct "
                        " --cellTag \n".format(cl)
                    )
                    continue
                else:
                    # per mil counts
                    out_file[3] *= 1.0 / (mil_reads_mapped)
            elif normUsing == "Mean":
                # divided by nCells
                out_file[3] *= 1.0 / (nCells)

            # out
            bg_out = "{}_{}.bedgraph".format(out_file_prefix, str(cl))

            out_file.to_csv(bg_out, sep="\t", index=False, header=False)
            os.remove(tmp_out)
            if format == "bigwig":
                bedGraphToBigWig(
                    chrom_names_and_size,
                    [bg_out],
                    "{}_{}.bw".format(out_file_prefix, str(cl)),
                )

    def writeBedGraph_worker(self, chrom, start, end, func_to_call, func_args, bed_regions_list=None):
        r"""Writes a bedgraph based on the read coverage per group of cells, indicated by cluster_info data frame.

        The given func is called to compute the desired bedgraph value
        using the funcArgs

        Parameters
        ----------
        chrom : str
            Chrom name
        start : int
            start coordinate
        end : int
            end coordinate
        func_to_call : str
            function name to be called to convert the list of coverages computed
            for each bam file at each position into a single value. An example
            is a function that takes the ratio between the coverage of two
            bam files.
        func_args : dict
            dict of arguments to pass to `func`.
        smoothLength : int
            Distance in bp for smoothing the coverage per tile.
        bed_regions_list: list
            List of tuples of the form (chrom, start, end)
            corresponding to bed regions to be processed.
            If not bed file was passed to the object constructor
            then this list is empty.

        Returns
        -------
        A list of [chromosome, start, end, temporary file], where the temporary file contains the bedgraph results for the region queried.

        Examples
        --------
        >>> test_path = os.path.dirname(os.path.abspath(__file__)) + "/test/test_data/"
        >>> bamFile1 = test_path +  "testA.bam"
        >>> bin_length = 50
        >>> number_of_samples = 0 # overruled by step_size
        >>> func_to_call = scaleCoverage
        >>> funcArgs = {'scaleFactor': 1.0}

        >>> c = WriteBedGraph([bamFile1], bin_length, number_of_samples, stepSize=50)
        >>> tempFile = c.writeBedGraph_worker( '3R', 0, 200, func_to_call, funcArgs)
        >>> f = open(tempFile[3], 'r')
        >>> f.readlines()
        ['3R\t0\t100\t0\n', '3R\t100\t200\t1\n']
        >>> f.close()
        >>> os.remove(tempFile[3])
        """
        if start > end:
            raise NameError("start position ({0}) bigger " "than end position ({1})".format(start, end))
        coverage, _, r = self.count_reads_in_region(chrom, start, end)

        ## get groups (clusters)
        cluster_info = self.clusterInfo
        clusters = cluster_info.cluster.unique().tolist()
        tempfilenames = dict.fromkeys(clusters)
        ## sum up tilecoverage group-wise
        for cl in clusters:
            if pd.isna(cl):
                continue
            _file = open(utilities.getTempFileName(suffix=".bg"), "w")
            previous_value = None
            line_string = "{}\t{}\t{}\t{:g}\n"
            cl_idx = cluster_info.index[pd.Series(cluster_info.cluster == cl)].tolist()

            for tileIndex in range(coverage.shape[0]):
                ## smoothing disabled for now
                tileCoverage = coverage[tileIndex, :]
                if self.skipZeroOverZero and np.sum(tileCoverage) == 0:
                    continue

                value = func_to_call(np.sum(tileCoverage[cl_idx]), func_args)

                if previous_value is None:
                    writeStart = start + tileIndex * self.binLength
                    writeEnd = min(writeStart + self.binLength, end)
                    previous_value = value

                elif previous_value == value:
                    writeEnd = min(writeEnd + self.binLength, end)

                elif previous_value != value:
                    if not np.isnan(previous_value):
                        _file.write(line_string.format(chrom, writeStart, writeEnd, previous_value))
                    previous_value = value
                    writeStart = writeEnd
                    writeEnd = min(writeStart + self.binLength, end)

            # write remaining value if not a nan
            if previous_value is not None and writeStart != end and not np.isnan(previous_value):
                _file.write(line_string.format(chrom, writeStart, end, previous_value))

            tempfilenames[cl] = _file.name
            _file.close()

        return chrom, start, end, tempfilenames



from itertools import compress


## get overlap of GTF object (deeptoolsintervals) with anndata object (from sincei)
## output: dict (region->gene mapping)
def get_gtf_adata_olaps(adata, gtf):
    r"""Get overlaps between adata and gtf

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    gtf : GTF
        GTF object

    Returns
    -------
    dict
        Dictionary with overlaps for each gene in adata.

    Examples
    --------
    >>> test = Tester()
    >>> gtf = GTF(test.gtfFile)
    >>> adata = sc.read_10x_mtx(test.input_matrix_dir, var_names='gene_symbols', cache=True)
    >>> olaps=get_gtf_adata_olaps(adata, gtf)
    >>> olaps['Gm37381']
    [('ENSMUSG00000064372', '+'), ('ENSMUSG00000064372', '-')]
    """
    var = adata.var
    olaps = dict.fromkeys(var.index)
    for i, key in enumerate(var.index):
        try:
            chrom, start, end = (
                var["chrom"][i],
                int(var["start"][i]),
                int(var["end"][i]),
            )
            ol = gtf.findOverlaps(chrom, start, end, includeStrand=True)
            if ol:
                genelist = [(x[2], x[5]) for x in ol]
                olaps[key] = genelist
        except ValueError:
            olaps[key] = None
            continue
    return olaps


## Search for bins by gene name, return either the first bin (promoter) or all overlapping bins
def get_bins_by_gene(dict, gene, firstBin=False):
    r"""
    Returns the bins for a given gene.

    Parameters
    ----------
    dict : dict
        Dictionary of bins and genes.
    gene : str
        Gene name.
    firstBin : bool
        If true, return only the first bin of the gene.

    Returns
    -------
    list
        List of bins.

    Examples
    --------
    >>> dict = {'chr1_1': [('gene1', '+'), ('gene2', '-')], 'chr1_2': [('gene1', '+')]}
    >>> get_bins_by_gene(dict, 'gene1')
    ['chr1_1', 'chr1_2']
    >>> get_bins_by_gene(dict, 'gene1', firstBin=True)
    'chr1_1'
    """
    klist = []
    for k, v in dict.items():
        if v:
            vlist = [x[0] for x in v]  # overlapping genes
            slist = [x[1] for x in v]  # overlapping gene strands
            match = [x.lower() == gene.lower() for x in vlist]
            if any(match):
                klist.append(k)
                # get strand of the gene
                strand = list(compress(slist, match))[0]
            else:
                strand = None

    # if firstBin is asked, sort the bins by start pos and
    # return only the firstBin by strand
    if klist and firstBin:
        spos = [x.split("_")[1] for x in klist]
        if strand == "+":
            first_bin = spos.index(min(spos))
        else:
            first_bin = spos.index(max(spos))
        return klist[first_bin]
    else:
        return klist



from sincei.scCountReads import *
from sincei import ReadCounter as countR
from sincei.Utilities import *

import os
import pandas as pd
import numpy as np
import numpy.testing as nt

ROOT = os.path.dirname(os.path.abspath(__file__)) + "/_data/"


def getCountReadsArgs(T):
    "return arg object based on input T"
    if T == "bins":
        TYPE = "bins -bs 10000"
    elif T == "gtf":
        TYPE = "features --BED {0}/Ogfrl.gtf".format(ROOT)
    elif T == "bed":
        TYPE = "features --BED {0}/test_regions.bed".format(ROOT)

    Args = (
        "{1} -b {0}/SL2-1.bam {0}/SL2-2.bam"
        " -bc {0}/test_barcodes.txt -ct BC -o None --smartLabels "
        "--region chr1:23365000:23385000".format(ROOT, TYPE).split()
    )
    args, newlabels = ParserCommon.validateInputs(parseArguments().parse_args(Args))

    return args, newlabels


def getCountReadsOutput(arg, dedup):
    """
    Setup the CR object based on testdata (2x bams, 5 barcodes) and input args
    """
    args, newlabels = getCountReadsArgs(arg)

    stepSize = args.binSize + args.distanceBetweenBins
    args.duplicateFilter = dedup

    c = countR.CountReadsPerBin(
        args.bamfiles,
        binLength=args.binSize,
        stepSize=stepSize,
        barcodes=args.barcodes,
        cellTag=args.cellTag,
        groupTag=args.groupTag,
        groupLabels=newlabels,
        numberOfSamples=None,
        region=args.region,
        bedFile=args.BED,
        duplicateFilter=args.duplicateFilter,
        zerosToNans=False,
        out_file_for_raw_data=None,
    )
    num_reads_per_bin, regionList = c.run(allArgs=args)

    return num_reads_per_bin, regionList


def getExpectedOutput(T, dedup):
    if T == "bins":
        regions = np.array(
            ["chr1_23360000_23370000::None", "chr1_23370000_23380000::None", "chr1_23380000_23385000::None"]
        )
        if dedup is None:
            counts = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [14.0, 0.0, 0.0, 32.0, 10.0, 2.0, 0.0, 22.0, 4.0, 3.0],
                    [0.0, 6.0, 4.0, 0.0, 0.0, 0.0, 6.0, 4.0, 0.0, 8.0],
                ]
            )
        if dedup == "start_bc_umi":
            counts = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [6.0, 0.0, 0.0, 8.0, 6.0, 2.0, 0.0, 5.0, 2.0, 2.0],
                    [0.0, 3.0, 2.0, 0.0, 0.0, 0.0, 3.0, 4.0, 0.0, 5.0],
                ]
            )

    elif T == "bed":
        regions = np.array(
            [
                "chr1_23365000_23377000::chr1:23365000-23377000",
                "chr1_23365000_23377000::chr1:23365000-23377000_r1",
                "chr1_23365000_23385000::chr1:23365000-23385000",
            ]
        )
        if dedup is None:
            counts = np.array(
                [
                    [0.0, 0.0, 0.0, 32.0, 10.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 32.0, 10.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                    [14.0, 6.0, 4.0, 32.0, 10.0, 2.0, 6.0, 26.0, 4.0, 10.0],
                ]
            )
        if dedup == "start_bc_umi":
            counts = np.array(
                [
                    [0.0, 0.0, 0.0, 8.0, 6.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 8.0, 6.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                    [6.0, 3.0, 2.0, 8.0, 6.0, 2.0, 3.0, 9.0, 2.0, 6.0],
                ]
            )

    elif T == "gtf":
        regions = np.array(
            [
                "chr1_23366423_23383175::ENSMUST00000027343.5",
                "chr1_23369723_23380801::ENSMUST00000186064.6",
                "chr1_23370267_23397541::ENSMUST00000188677.1",
            ]
        )
        if dedup is None:
            counts = np.array(
                [
                    [14.0, 6.0, 0.0, 32.0, 10.0, 2.0, 6.0, 26.0, 4.0, 10.0],
                    [14.0, 6.0, 0.0, 32.0, 10.0, 2.0, 0.0, 26.0, 4.0, 10.0],
                    [14.0, 6.0, 4.0, 32.0, 4.0, 2.0, 6.0, 26.0, 4.0, 10.0],
                ]
            )
        if dedup == "start_bc_umi":
            counts = np.array(
                [
                    [6.0, 3.0, 0.0, 8.0, 6.0, 2.0, 3.0, 9.0, 2.0, 6.0],
                    [6.0, 3.0, 0.0, 8.0, 6.0, 2.0, 0.0, 9.0, 2.0, 6.0],
                    [6.0, 3.0, 2.0, 8.0, 4.0, 2.0, 3.0, 9.0, 2.0, 6.0],
                ]
            )

    return counts, regions


def testCountReads_bins_default():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("bins", None)
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("bins", None)
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)


def testCountReads_bins_dedup():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("bins", "start_bc_umi")
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("bins", "start_bc_umi")
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)


def testCountReads_bed_default():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("bed", None)
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("bed", None)
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)


def testCountReads_bed_dedup():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("bed", "start_bc_umi")
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("bed", "start_bc_umi")
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)


def testCountReads_gtf_default():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("gtf", None)
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("gtf", None)
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)


def testCountReads_gtf_dedup():
    # Expected output
    valid_counts, valid_regions = getExpectedOutput("gtf", "start_bc_umi")
    # Actual output
    observed_counts, observed_regions = getCountReadsOutput("gtf", "start_bc_umi")
    # Test
    nt.assert_array_equal(valid_regions, observed_regions)
    nt.assert_array_equal(valid_counts, observed_counts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import numpy as np
import pandas as pd
from deeptools.plotFingerprint import getSyntheticJSD

from matplotlib.pyplot import plot
import matplotlib

matplotlib.use("Agg")
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["svg.fonttype"] = "none"
# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei import ReadCounter as countR
from sincei import ParserCommon

## plot KDE of JSD values
from numpy import array, linspace
from sklearn import neighbors

old_settings = np.seterr(all="ignore")
MAXLEN = 10000000


def get_args():
    parser = argparse.ArgumentParser(add_help=False, conflict_handler="resolve")
    optional = parser.add_argument_group("Optional arguments")

    optional.add_argument(
        "--numberOfSamples",
        "-n",
        help="The number of bins that are sampled from the genome, "
        "for which the overlapping number of reads is computed. (Default: %(default)s)",
        default=1e5,
        type=int,
    )

    optional.add_argument(
        "--skipZeros",
        help="If set, then regions with zero overlapping reads"
        "for *all* given BAM files are ignored. This "
        "will result in a reduced number of read "
        "counts than that specified in --numberOfSamples",
        action="store_true",
    )

    return parser


def parseArguments(args=None):
    io_args = ParserCommon.inputOutputOptions(
        opts=["bamfiles", "barcodes", "outFile"],
        requiredOpts=["bamfiles", "barcodes", "outFile"],
    )
    bam_args = ParserCommon.bamOptions(suppress_args=["region", "distanceBetweenBins"], default_opts={"binSize": 10000})
    read_args = ParserCommon.readOptions(suppress_args=["filterRNAstrand", "extendReads", "centerReads"])
    filter_args = ParserCommon.filterOptions(
        suppress_args=[
            "motifFilter",
            "genome2bit",
            "GCcontentFilter",
            "minAlignedFraction",
        ]
    )
    other_args = ParserCommon.otherOptions()

    parser = argparse.ArgumentParser(
        parents=[io_args, bam_args, read_args, filter_args, get_args(), other_args],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="This tool samples regions in the genome from BAM files "
        "and compares the cumulative read coverages for each cell on those regions. "
        "to a synthetic cell with poisson distributed reads using Jansson Shannon Distance. "
        "Cells with high enrichment of signals show a higher JSD compared to cells whose signal "
        "is homogenously distrubuted.",
        conflict_handler="resolve",
        usage="An example usage is: plotFingerprint -b treatment.bam control.bam " "-plot fingerprint.png",
        add_help=False,
    )

    return parser


def main(args=None):
    args = ParserCommon.process_args(parseArguments().parse_args(args))
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")

    ## read the barcode file
    with open(args.barcodes, "r") as f:
        barcodes = f.read().splitlines()
    f.close()

    ## Count
    c = countR.CountReadsPerBin(
        args.bamfiles,
        binLength=args.binSize,
        numberOfSamples=args.numberOfSamples,
        barcodes=barcodes,
        cellTag=args.cellTag,
        motifFilter=None,
        genome2bit=None,
        GCcontentFilter=None,
        blackListFileName=args.blackListFileName,
        numberOfProcessors=args.numberOfProcessors,
        verbose=args.verbose,
        region=None,
        bedFile=None,
        extendReads=args.extendReads,
        minMappingQuality=args.minMappingQuality,
        duplicateFilter=args.duplicateFilter,
        center_read=False,
        samFlag_include=args.samFlagInclude,
        samFlag_exclude=args.samFlagExclude,
        minFragmentLength=args.minFragmentLength,
        maxFragmentLength=args.maxFragmentLength,
        zerosToNans=False,
        sumCoveragePerBin=True,
    )

    num_reads_per_bin, _ = c.run(allArgs=None)

    if num_reads_per_bin.sum() == 0:
        import sys

        sys.stderr.write(
            "\nNo reads were found in {} regions sampled. Check that the\n"
            "min mapping quality is not overly high and that the \n"
            "chromosome names between bam files are consistant.\n"
            "For small genomes, decrease the --numberOfSamples.\n"
            "\n".format(num_reads_per_bin.shape[0])
        )
        exit(1)

    if args.skipZeros:
        num_reads_per_bin = countR.remove_row_of_zeros(num_reads_per_bin)

    total = len(num_reads_per_bin[:, 0])
    x = np.arange(total).astype("float") / total  # normalize from 0 to 1

    jsd_all = []
    for i in range(0, num_reads_per_bin.shape[1]):
        jsd_all.append(getSyntheticJSD(num_reads_per_bin[:, i]))

    ## create colnames (sampleLabel+barcode)
    newlabels = ["{}_{}".format(a, b) for a in args.labels for b in barcodes]
    df = pd.DataFrame({"cell": newlabels, "jsd": jsd_all})
    df.to_csv(args.outFile, sep="\t")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import re
import argparse
import pandas as pd
import scanpy as sc

# logs
import warnings
import logging

logger = logging.getLogger()
warnings.simplefilter(action="ignore", category=FutureWarning)

from sincei import ParserCommon
from sincei.Utilities import gini


### ------ Functions ------
def filter_adata(
    adata,
    filter_region_dict=None,
    filter_cell_dict=None,
    bad_chrom=None,
    bad_cells=None,
):
    # 1. regions
    if bad_chrom:
        adata = adata[:, ~adata.var.chrom.isin(bad_chrom)]
    if filter_region_dict:
        for key in filter_region_dict.keys():
            adata = adata[
                :,
                (adata.var[key] >= filter_region_dict[key][0]) & (adata.var[key] <= filter_region_dict[key][1]),
            ]

    # 2. Cells
    if bad_cells:
        adata = adata[~adata.obs.index.isin(bad_cells)]
    if filter_cell_dict:
        for key in filter_cell_dict.keys():
            adata = adata[
                (adata.obs[key] >= filter_cell_dict[key][0]) & (adata.obs[key] <= filter_cell_dict[key][1]),
                :,
            ]

    return adata


"""
def make_plots(adata, fname=None):
    # plotting
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('Agg')
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['svg.fonttype'] = 'none'
    import seaborn as sns# seaborn colormaps conflicts with deeptools colormaps

    ## filtering of regions
    plt.figure()
    pl1=sns.violinplot(data=adata.var, x='total_counts', y='chrom')
    plt.figure()
    pl2=sns.distplot(np.log10(adata.var['total_counts']+1))
    plt.figure()
    pl3=sns.distplot(adata.var['mean_counts'])
    plt.figure()
    pl4=sns.distplot(adata.var['fraction_cells_with_signal'])
    ## cells
    plt.figure()
    pl5=sns.scatterplot(data=adata.obs, x='total_counts', y='pct_counts_in_top_100_genes')
    pl5.set(xscale="log")
    plt.figure()
    pl6=sns.distplot(adata.obs['fraction_regions_with_signal'])
    plt.figure()
    pl7=sns.scatterplot(data=adata.obs, x='total_counts', y='fraction_regions_with_signal')

    plist=[pl1, pl2, pl3, pl4, pl5, pl6, pl7]
    if fname:
        with PdfPages(fname) as pp:
            for plot in plist:
                pp.savefig(plot.figure)
    return plist

    general.add_argument('--outPlot', '-op',
                         type=str,
                         help='The output plot file. This describes the distribution of filtering metrics pre and post filtering')
    if args.outPlot:
        make_plots(adata, fname=args.outPlot)
        if cellfilter or regionfilter or badcells or badchrom:
            make_plots(adata_filt, fname=args.outPlot+".filtered.pdf")
"""


def parseArguments():
    io_args = ParserCommon.inputOutputOptions(opts=["h5adfile", "outFile"])
    plot_args = ParserCommon.plotOptions()
    other_args = ParserCommon.otherOptions()
    parser = argparse.ArgumentParser(
        parents=[io_args, get_args(), plot_args, other_args],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
        This tool performs calculates multiple quality controls metrics on the input .h5ad file (output of scCountReads)
        and (optionally) filters the input file based on filterCellArgs/filterRegionArgs. The output is either an
        updated .h5ad object (if filtering is requested) or the filtering metrics (--outMetrics) and plots (--outPlot).
        """,
        usage="Example usage: scCountQC -i cellCounts.h5ad -om qc_metrics.tsv > log.txt",
        add_help=False,
    )

    return parser


def get_args():
    parser = argparse.ArgumentParser(add_help=False)

    general = parser.add_argument_group("Filtering arguments")
    general.add_argument(
        "--describe",
        "-d",
        action="store_true",
        help="Print a list of cell and region metrics available for QC/filtering.",
    )

    general.add_argument(
        "--outMetrics",
        "-om",
        type=str,
        help="Prefix of the output file with calculated QC metrics. If given, the cell metrics are printed in "
        "<prefix>.cell.tsv and region metrics as <prefix>.region.tsv",
    )

    general.add_argument(
        "--filterCellArgs",
        "-fc",
        type=str,
        help='List of arguments to filter cells. The format is "arg_name: minvalue, maxvalue; arg_name: minvalue, maxvalue; ...." '
        "where arg_name is the QC metric for cells present in the input h5ad file. In order to view all available "
        'cell filtering metrics, run scCountFilter with "--describe". The two arguments are supplied (minvalue, maxvalue) '
        "they are used as lower and upper bounds to filter cells. Make sure they are float/integer numbers.",
    )

    general.add_argument(
        "--filterRegionArgs",
        "-fr",
        type=str,
        help='List of arguments to filter regions. The format is "arg_name: minvalue, maxvalue; arg_name: minvalue; ...." '
        "where arg_name is the QC metric for regions present in the input h5ad file. In order to view all available "
        'cell filtering metrics, run scCountFilter with "--describe". The two arguments are supplied (minvalue, maxvalue) '
        "they are used as lower and upper bounds to filter cells. Make sure they are float/integer numbers.",
    )

    general.add_argument(
        "--cell_blacklist",
        "-cb",
        default=None,
        type=argparse.FileType("r"),
        help="A list of barcodes to be included for the clustering. The barcodes "
        "(along with sample labels) must be present in the input object.",
    )

    general.add_argument(
        "--chrom_blacklist",
        "-chb",
        default=None,
        type=str,
        help="A comma separated list of chromosomes to exclude. eg. chrM, chrUn",
    )

    return parser


def main(args=None):
    args = parseArguments().parse_args(args)
    if not args.verbose:
        logger.setLevel(logging.CRITICAL)
        warnings.filterwarnings("ignore")
    try:
        adata = sc.read_h5ad(args.input)  # , obs_names="obs_names", var_names="var_names")
    except:
        sys.stderr.write("\n Error: Input file can not be read (doesn't appear to be a valid anndata object) \n")
        exit()
    ## add QC stats to the anndata object
    # 1. scanpy metrics # fraction of regions/genes with signal are included in the metrics (pct_dropouts/n_genes_by_counts)
    try:
        sc.pp.calculate_qc_metrics(adata, inplace=True)
    except IndexError:  # not enough genes/regions
        sys.stderr.write("\n Error: Too few regions in the input file to perform QC \n")
        exit()

    # 2. sincei metrics
    gini_list = [gini(i, adata.X) for i in range(adata.shape[0])]
    adata.obs["gini_coefficient"] = gini_list

    if args.outMetrics:
        mat = re.sub(".txt|.tsv|.csv", "", args.outMetrics)
        obs_tsv = mat + ".cells.tsv"
        var_tsv = mat + ".regions.tsv"
        adata.obs.to_csv(obs_tsv, sep="\t", index_label="Cell_ID")
        adata.var.to_csv(var_tsv, sep="\t", index_label="Feature_ID")

    # if --describe is asked, only print the numeric vars and obs columns
    if args.describe:
        is_num_col = [(pd.api.types.is_float_dtype(x) | pd.api.types.is_integer_dtype(x)) for x in adata.obs.dtypes]
        cols = adata.obs.loc[:, is_num_col]
        sys.stdout.write("\n Cell metrics: \n")
        sys.stdout.write("Total cells: {} \n".format(adata.shape[0]))
        print(pd.DataFrame({"min": cols.min(), "max": cols.max()}))

        is_num_row = [(pd.api.types.is_float_dtype(x) | pd.api.types.is_integer_dtype(x)) for x in adata.var.dtypes]
        rows = adata.var.loc[:, is_num_row]
        sys.stdout.write("\n Feature metrics: \n")
        sys.stdout.write("Total features: {} \n".format(adata.shape[1]))
        print(pd.DataFrame({"min": rows.min(), "max": rows.max()}))
        exit()

    if args.filterCellArgs:
        cellfilter = dict()
        for x in args.filterCellArgs.strip().split(";"):
            key = x.strip().split(":")[0]
            v = x.strip().split(":")[1]
            value = [float(x) for x in v.strip().split(",")]
            cellfilter[key] = value
    else:
        cellfilter = None
    if args.filterRegionArgs:
        regionfilter = dict()
        for x in args.filterRegionArgs.strip().split(";"):
            key = x.strip().split(":")[0]
            v = x.strip().split(":")[1]
            value = [float(x) for x in v.strip().split(",")]
            regionfilter[key] = value
    else:
        regionfilter = None

    if args.cell_blacklist:
        ## read the barcode file
        with open(args.cell_blacklist, "r") as f:
            badcells = f.read().splitlines()
        f.close()
    else:
        badcells = None

    if args.chrom_blacklist:
        badchrom = args.chrom_blacklist.strip().split(",")
    else:
        badchrom = None

    if cellfilter or regionfilter or badcells or badchrom:
        sys.stdout.write("Applying filters \n")
        adata_filt = filter_adata(
            adata,
            filter_region_dict=regionfilter,
            filter_cell_dict=cellfilter,
            bad_chrom=badchrom,
            bad_cells=badcells,
        )
        sys.stdout.write("Remaining cells: {} \n".format(adata_filt.shape[0]))
        sys.stdout.write("Remaining features: {} \n".format(adata_filt.shape[1]))
        adata_filt.write_h5ad(args.outFile)

    return 0
